import java.util.*;
import java.io.*;
import java.net.*;

/**
  * The Game Class is the central core of the the Sokoban game. The Game
  * class is a link between the user interface, the networking and
  * the map.
  *
  * @author  Antony George
  * @date    11/28/01
  * @version 1.4
  *
  */
//Version History
//  11/23/01 Antony George added algorithm pseudocode
//  11/28/01 Kevin Blomseth edited method headers and added object headers
//  12/01/01 Joshua Thompson added java code
//  01/11/02 Joshua Thompson added CustomLevelOrder support
//  01/23/02 Kamil Baranowski added java code pauseTimer(), restartMap(),
//           instantReplay(), setMultiplayer(), getTime(), endGame(),
//           & undoMove().
//  01/28/02 Joshua Thompson added java code for method setMapObserver() and
//			 comments for method getTime()
//  01/29/02 Kamil Baranowski added java code for postDisconnect()
//  and postMove()
//  02/28/02 Kevin Blomseth added postChat()

public class Game
{
    /**
     * Instantiation of the server
     */
    private Server serv;

    /**
     * Instantiation of the map class to store the data
     * of the level.
     */
    private Map map;

    /**
     * The name of the user in the current game.
     */
    private String userName;

    /**
     * The robot number of the usre in the current game.
     */
    private int robotNum;

    /**
     * The user defined order of user created levels.
     */
    private CustomLevelOrder customLevelOrder;

    /**
     * Instantiation of the client
     */
    private Client client;

    /**
     * The list of the 15 highest scores.
     */
    private HighScoresList highScores;

    /**
     * Says whether or not this game is acting as a host.
     */
    private boolean hosting;

    /**
     * Says whether or not this game is a multiplayer game.
     */
    private boolean multiplayer;

    /**
     * The link to the Sokoban user interface.
     */
    protected SGui gui;

    /**
     * The link to the multiplayer interface.
     */
    private MultiplayerDialog mui;

    /**
     * The total cummulative score for the user.
     */
    private int totalScore;

    /**
     * The flag signifying if the game should return 1 as the time or not.
     */
    private boolean debugMode;

    /**
     * The number of restart votes cast (for use in multiplayer)
     */
    private int restartVotes = 0;

    /**
     * The number of change level votes cast (for use in multiplayer)
     */
    private int changeVotes = 0;

    /**
     * The flag signifying if the game is currently active or not.
     */
    private boolean active;

    /**
     * Constructor - initializes the game class
     */
    public Game (SGui sokobanGUI)
    {
        // SET gui = sokobanGUI
        this.gui = sokobanGUI;
        // SET totalScore = 0
        totalScore = 0;
        // INIT customLevelOrder
        customLevelOrder = new CustomLevelOrder();
        // SET multiplayer = false
        multiplayer = false;
        // SET hosting = false
        hosting = false;
        // SET debugMode = false
        debugMode = false;
        // SET default userName = "Player"
        userName = "Player";
        // SET robotNum = 0
        robotNum = 0;
        // LOAD scores
        highScores = new HighScoresList(this);
        highScores.loadHighScores();
        // INIT map
        map = new Map(this);
        // SET active = true
        active = true;
        // ADD new Robot with userName and robotNum to the map
        map.addRobot(new Robot(userName, robotNum));
        // OPEN level list
        try
        {
            BufferedReader in = new BufferedReader(new FileReader(Constants.levelList));
            // INIT new String level
            String level;
            // INIT new File levelFile
            File levelFile;
            // WHILE file is ready LOOP
            while (in.ready())
            {
                // READ location from file
                level = in.readLine();
                levelFile = new File(Constants.levelPath+level+Constants.levelExt);
                if (levelFile.exists())
                {
                    customLevelOrder.addLevel(level);
                }
                // ENDWHILE
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    /**
     * This function tells the system that level has been completed and to therefore
     * proceed to next level.
     *
     * <BR><B>Precondition:</B>All conditions for completing a certain level have been met
     * <BR><B>Postcondition:</B>The system proceeds to the next level to be played by the
     * user
     */
    public void levelCompleted()
    {
        // INIT new int score = map.calculateScore() passing robotNum
        int score = map.calculateScore(robotNum);
        // SET totalScore = totalScore + score
        totalScore += score;
        // CALL method pauseTimer()
        pauseTimer();
        // IF debugMode THEN
        if(debugMode)
        {
            // SET score = 20
            score = 20;
        }
        // CALL method gui.setLevelFinishedDisplay() passing score
        gui.setLevelFinishedDisplay(score);
    }

    /**
     * This function allows the user to save his progress in the game
     *
     * <BR><B>Precondition:</B>User is engaged in a game
     * <BR><B>Postcondition:</B>The game is saved
     *
     * @param saveSlot the saved game slot to save to
     * @param description the description entered by the user describing the current
     * saved game.
     */
    public void saveGame(int saveSlot, String description)
    {
        // APPEND saveSlot number to the end of savegame filename String
        // OPEN savegame file
        // WRITE level data to file
        // CLOSE savegame file
        try
        {
            // CHECK if saveDir exists and create if it doesn't
            File saveDir = new File(Constants.savePath);
            if (!saveDir.exists())
            {
                saveDir.mkdir();
            }
            // INIT robot = getLocalRobot()
            Robot robot = getLocalRobot();
            String file = Constants.savePath + "GAME" + saveSlot + Constants.saveExt;
            PrintWriter outFile = new PrintWriter(new FileWriter(file));
            //PRINT user entered description to file
            outFile.println(description);
            //PRINT user entered level name to file
            outFile.println(customLevelOrder.getCurrentLevel());
            //PRINT cumulative score to file
            outFile.println(""+robot.getScore());
            //PRINT current time for level to file
            outFile.println(""+getTime());
            //GET moves from robot
            Move[] moves = robot.getMoveTracker();
            //FOR vector length amount of times
            for (int index = 0; index < moves.length; index++)
            {
                //PRINT each move to save file
                outFile.println(moves[index].getDirection().toString());
                //ENDFOR
            }
            //CLOSE output file



            outFile.close();
        }
        catch (IOException exc)
        {
            exc.printStackTrace();
        }
    }

    /**
     * This function allows the user to load a game from a previously saved game
     *
     * <BR><B>Precondition:</B>none
     * <BR><B>Postcondition:</B>The previously played game is loaded and the
     * map is set to the saved state
     *
     * @param loadSlot the saved game slot to be loaded from
     */
    public void loadGame(int loadSlot)
    {
        // APPEND loadSlot number to the end of savegame filename String
        // OPEN savegame file
        // WHILE not end of file LOOP
        // READ level data from file
        // CLOSE savegame file
        // INIT new Map
        // INIT new Robot with userName, color
        // CALL addRobot() within Map passing the Robot
        try
        {
            map.loading = true;
            String line;
            String filename = Constants.savePath + "GAME" + loadSlot + Constants.saveExt;
            File loadFile = new File(filename);
            if (loadFile.exists())
            {
                BufferedReader inFile = new BufferedReader(new FileReader(loadFile));
                // READ the first line to skip the description
                inFile.readLine();
                String levelName = inFile.readLine();
                startMap(levelName);
                this.totalScore = Integer.parseInt(inFile.readLine());
                int time = Integer.parseInt(inFile.readLine());
                map.setTimer(time);
                Robot robot = getLocalRobot();
                while ((line = inFile.readLine()) != null)
                {
                    Direction dir = null;
                    if (line.equals(Direction.UP.toString()))
                    {
                        dir = Direction.UP;
                    }
                    else if (line.equals(Direction.DOWN.toString()))
                    {
                        dir = Direction.DOWN;
                    }
                    else if (line.equals(Direction.LEFT.toString()))
                    {
                        dir = Direction.LEFT;
                    }
                    else if (line.equals(Direction.RIGHT.toString()))
                    {
                        dir = Direction.RIGHT;
                    }
                    map.moveRobot(robot.getRobotNumber(), dir);
                }
                inFile.close();
            }
            map.loading=false;
        }
        catch (IOException exc)
        {
            exc.printStackTrace();
        }
    }

    /**
     * This function initializes the map, adds the robot(s) to the map and
     * then loads the level to be played.
     *
     * @param levelName the name of the initial level to play
     * <BR><B>Precondition:</B>none
     * <BR><B>Postcondition:</B>The level is loaded and the game is started
     */
    public void startMap(String levelName)
    {
        // CALL method gui.startGame() passing value returned from isMultiplayer()
        gui.startGame(isMultiplayer());
        // CALL method Game.updateDisplay()
        updateDisplay();
        if (hosting)
        {
            serv.stopListening();
        }
        // CALL method Game.changeLevel() passing levelName
        doChangeLevel(levelName);
        // UPDATE display
        updateDisplay();
    }

    /**
     * Processes a request to change the current level being played.
     *
     * @param level the name of the level to change to
     * <br><b>preconditions:</b> the game is started
     * <br><b>postconditions:</b> the level passed is loaded if single player,
     * or a request to change to the level passed is sent to the other users
     * if multiplayer.
     */
    public void changeLevel(String level)
    {
        if (multiplayer && map.getRobots().length > 1)
        {
            // IF hosting THEN
            if (hosting)
            {
                Robot robot = getLocalRobot();
                // CALL method sendRestartMap() from Server serv
                serv.sendChangeLevelRequest(level, robot.getRobotNumber());
                robot.setChangeLevelFlag(true);
                changeVotes++;
            }
            // ELSE
            else
            {
                // CALL method sendRestartMap() from Client client
                client.sendChangeLevelRequest(level);
                // ENDIF
            }
        }
        else
        {
            doChangeLevel(level);
        }
    }

    /**
     * Tells the map class to load the level passed.
     *
     * @param level    String value associated with name of the level
     * <br><b>Pre-conditions: none </b>
     * <br><b>Post-conditions: level is sent to changeLevel within the Map class
     * <br>
     */
    public void doChangeLevel(String level)
    {
        // CALL method changeLevel() within Map
        map.changeLevel(level);
        // CALL method map.restartLevel()
        map.restartLevel();
        // CALL method map.startTimer()
        map.startTimer();
        // CALL method gui.setMapVisible() to set the map to be displayed
        gui.setMapVisible();
        // CALL method gui.setLevelName(level) to set level name display
        gui.setLevelName(level);
        // CALL method customLevelOrder.setCurrentLevel() passing level
        customLevelOrder.setCurrentLevel(level);
        // IF multiplayer THEN
        if (multiplayer)
        {
            // IF hosting THEN
            if (hosting)
            {
                // CALL method serv.sendMap() passing level
                serv.sendMap(level);
                // ENDIF
            }
            // HIDE the multiplayer window if it's still visible



            mui.dispose();
            // ENDIF
        }
    }

    /**
     * Allows the user to agree or disagree to a change level request.  If
     * hosting, the change request is sent to the other users
     *
     * @param levelName the name of the level to change to
     * @param nRobot the robot number requesting the change
     *
     * <br><b>precondition:</b> a multiplayer game is in progress
     * <br><b>postcondition:</b> the user has agreed or disagreed to the change
     * level request
     */
    public void requestChangeLevel(String levelName, int nRobot)
    {
        // GET the robot from the map
        Robot robot = map.getRobot(nRobot);
        // IF hosting THEN
        if (hosting)
        {
            // SET robot's change level flag = true
            robot.setChangeLevelFlag(true);
            // CALL method serv.sendChangeLevelRequest()
            serv.sendChangeLevelRequest(levelName, nRobot);
            // INCREMENT changeVotes
            changeVotes++;
            // POST server chat message of request
            // ENDIF
        }
        // INIT new boolean accepted = mui.showChangeLevelRequestDialog()



        boolean accepted = mui.showChangeLevelRequestDialog(robot.getUserName(), levelName);
        // IF hosting THEN
        if (hosting)
        {
            // SET robot = getLocalRobot()
            robot = getLocalRobot();
            // CALL method postAcceptedChangeLevel
            postAcceptedChangeLevel(levelName, robot.getRobotNumber(), accepted);
            // POST server chat message of accept
        }
        // ELSE



        else
        {
            // CALL client.sendChangeLevelAccept
            client.sendChangeLevelAccept(accepted, levelName);
            // ENDIF
        }
    }

    /**
     * Sets a robot's change level flag to the accepted boolean passed.
     *
     * @param level the name of the level to change to
     * @param robotNum the robotnum of the user accepting or declining
     * @param accepted true if the user accepts, false if not
     *
     * <br><b>preconditions:</b> a multiplayer game is in progress
     * <br><b>postconditions:</b> the user with robotNum has selected to accept or
     * decline the request for a level change
     */
    public void postAcceptedChangeLevel(String level, int robotNum, boolean accepted)
    {
        // IF multiplayer and hosting THEN
        if (multiplayer && hosting)
        {
            // SET change flag of robot[robotNum] = accepted
            map.getRobot(robotNum).setChangeLevelFlag(accepted);
            // INCREMENT changeVotes()
            changeVotes++;
            // CALL method checkChangeLevelVotes() passing level
            checkChangeLevelVotes(level);
            // ENDIF
        }
    }

    /**
     * Checks to see if the change level vote has completed.
     *
     * @param level the name of the level to change to
     *
     * <br><b>preconditions:</b> a change level vote is in progress
     * <br><b>postconditions:</b> if the vote is complete and it was unanimous
     * in favor then the level is changed.  otherwise a vote failed message
     * is sent to the chat
     */
    protected void checkChangeLevelVotes(String level)
    {
        // INIT new boolean ready = true
        boolean ready = true;
        // INIT new int index = 0
        int index = 0;
        // INIT robots = map.getRobots()
        Robot[] robots = map.getRobots();
        // WHILE ready and index < length of robots DO
        while (ready && index < robots.length)
        {
            // SET ready = robots[index] change level flag
            ready = robots[index].getChangeLevelFlag();
            // INCREMENT index
            index++;
            // ENDWHILE
        }
        // IF ready THEN



        if (ready)
        {
            // CALL method doChangeLevel() passing level
            doChangeLevel(level);
            // ENDIF
        }
        // changeVotes >= robots.length



        if (changeVotes >= robots.length)
        {
            // FOR bot = 0 to robots.length
            for (int bot = 0; bot < robots.length; bot++)
            {
                // SET change level flag of robot[bot] = false
                robots[bot].setChangeLevelFlag(false);
                // ENDFOR
            }
            // SET changeVotes = 0



            changeVotes = 0;
            // ENDIF
        }
    }


    /**
     * This function displays the chat messages sent by the users.
     *
     * @param user     The name of the user sending the message
     * @param msg      The message sent by user
     * <BR><B>Precondition:</B>User is engaged in a multiplayer game
     * <BR><B>Postcondition:</B>The message is displayed in the
     * chat window for all users to read
     */
    public void postChat(String user, String msg)
    {
        // IF hosting THEN
        if (hosting)
        {
            // CALL sendChat() within Server serv passing user and msg
            serv.sendChat(user, msg);
            // ELSE
        }
        else
        {
            // CALL sendChat() within Client client passing user and msg
            client.sendChat(user, msg);
        }
        // ENDIF
        // send message to GUI to display user and msg in GUI chat box
        gui.addText(user+": "+msg);
        //ENDIF
    }

    /**
     * Informs other users connected to a server of a client disconnecting
     * and removes the client's robot from the map
     *
     * @param robotNum The robot number of the user disconnecting
     * <BR><B>Precondition:</B>User is engaged in a multiplayer game
     * <BR><B>Postcondition:</B>The robot number of the user disconnecting is
     * removed from the map.
     */
    public void postDisconnect(int robotNum)
    {
        // CALL mui.removeUser()
        map.removeRobot(robotNum);
        if (multiplayer)
        {
            // IF hosting THEN
            if (hosting)
            {
                // CALL sendDisconnect() within Server serv passing robotNum
                serv.sendDisconnect(robotNum);
            }
            // ELSE if robotNum = this.robotNum
            else if (robotNum == this.robotNum)
            {
                gui.endGame();
            }
            mui.updateUserList(getRobots());
        }
    }

    /**
     * This function posts the move made by the user on the map.
     *
     * @param robotNum     The id of the robot being moved
     * @param dir    The direction in which the robot moved
     * <BR><B>Precondition:</B>User is engaged in a multiplayer game
        * <BR><B>Postcondition:</B>The direction in which the particular
        * robot moved is displayed to all users
     */
    public void postMove(int robotNum, Direction dir)
    {
        if (multiplayer)
        {
            // IF hosting THEN
            if (hosting)
            {
                // CALL class Server method sendMove with robotNum and direction
                serv.sendMove(robotNum, dir);
            }
            // ELSE
            else
            {
                // CALL method Client.sendMove()
                client.sendMove(robotNum, dir);
                // ENDIF
            }
            // ENDIF
        }
        // CALL class Map method moveRobot



        map.moveRobot(robotNum, dir);
    }

    /**
     * Sends undoMove message to the Map class.  Sends the message to the server
     * if the user is hosting and to the client if the user is not.
     *
     * @param robotNum     The id of the robot undoing a move
     * <BR><B>Precondition:</B>User is engaged in a multiplayer game
     * <BR><B>Postcondition:</B>Returns the map to the state it was in before
     * the most recent move was executed.
     */
    public void undoMove(int robotNum)
    {
        if (multiplayer)
        {
            // IF hosting THEN
            if (hosting)
            {
                // CALL method sendUndo() from Server serv
                serv.sendUndo(robotNum);
            }
            // ELSE
            else
            {
                // CALL method sendUndo() from Client client
                client.sendUndo(robotNum);
                // ENDIF
            }
        }
        // CALL class Map method undoMove



        map.undoMove(robotNum);
    }

    /**
     * Adds a user to the current game.
     *
     * @param userName the name of the new user
     * @param robotNum the robot number of the new user
     *
     * <b>precondition:</b> there is a multiplayer game running
     * <b>postcondition:</b> the new user has been added to the game
     */
    public void postNewUser(String userName, int robotNum)
    {
        // INIT new Robot robot with userName and robotNum
        Robot robot = new Robot(userName, robotNum);
        // ADD robot to current map
        map.addRobot(robot);
        // IF multiplayer THEN
        if (multiplayer)
        {
            // IF hosting THEN
            if (hosting)
            {
                // CALL method serv.addUser()
                serv.addUser(userName, robotNum);
            }
            // ELSE IF (robotNum = client.getRobotNumber())
            else if (robotNum == client.getRobotNumber())
            {
                // SET this.robotNum = robotNum
                this.robotNum = robotNum;
                // ENDIF
            }
            // CALL method mui.updateUserList() passing getRobots()



            mui.updateUserList(getRobots());
            // ENDIF
        }
    }

    /**
     * <p>Returns the current Map.</p>
     *
     * @return an array of Strings representing the current map
     */
    public String[] getMapString()
    {
        // IF map not equal to null THEN
        if (map != null)
        {
            // CALL method toStringArray() within the Map class
            // RETURN array returned from toStringArray() representing the map
            return map.toStringArray();
        }
        // ELSE
        else
        {
            // RETURN null
            return null;
            // END IF
        }
    }

    /**
        * <p>Returns the current Map.</p>
        *
        * @return a 2-dimensional array of Squares representing the current map
        */
    public Square[][] getMap()
    {
        // IF map not equal to null THEN
        if (map != null)
        {
            // CALL method getMap() within the Map class
            // RETURN array returned from getMap() representing the map
            return map.getMap();
        }
        // ELSE
        else
        {
            // RETURN null
            return null;
            // END IF
        }
    }

    /**
     * <p>Returns the robots currently in the map</p>
     *
     * @return an array of Robots
     */
    public Robot[] getRobots()
    {
        // CALL method getRobots() within the Map class
        // RETURN array returned from getRobots()
        return map.getRobots();
    }

    /**
     * Sends pauseTimer message to the Map class
     *
     * <BR><B>Precondition:</B> The game is not multiplayer
     * <BR><B>Postcondition:</B> The timer is paused.
     */
    public void pauseTimer()
    {
        // IF not multiplayer THEN
        if (!multiplayer)
        {
            // CALL method Map.pauseTimer()
            map.pauseTimer();
            // ENDIF
        }
    }

    /**
     * Sends startTimer message to the Map class
     *
     * <BR><B>Precondition:</B> The game is not multiplayer
     * <BR><B>Postcondition:</B> The timer is started.
     */
    public void startTimer()
    {
        // IF not multiplayer THEN
        if (!multiplayer)
        {
            // CALL method Map.startTimer()
            map.startTimer();
            // ENDIF
        }
    }

    /**
     * Returns true if the map is currently paused, false otherwise.
     *
     * @return returns whether the map is paused or not
     */
    public boolean isPaused()
    {
        // IF not multiplayer THEN
        if (!multiplayer)
        {
            // CALL method Map.isPaused()
            return map.isPaused();
            // ENDIF
        }
        // RETURN false if multiplayer



        return false;
    }

    /**
     * Sends restartMap message to the Map class
     *
     * <BR><B>Precondition:</B> None.
     * <BR><B>Postcondition:</B> The game has been restarted.
     */
    public void restartMap()
    {
        // IF multiplayer and number of robots is > 1 THEN
        if (multiplayer && map.getRobots().length > 1)
        {
            // IF hosting THEN
            if (hosting)
            {
                // INIT robot = getLocalRobot()
                Robot robot = getLocalRobot();
                // CALL method sendRestartMap() from Server serv
                serv.sendRestartRequest(robot.getRobotNumber());
                // SET robot's restart flag true
                robot.setRestartFlag(true);
                // INCREMENT restartVotes
                restartVotes++;
            }
            // ELSE
            else
            {
                // CALL method sendRestartMap() from Client client
                client.sendRestartRequest();
                // ENDIF
            }
        }
        // ELSE



        else
        {
            // CALL method doRestartMap()
            doRestartMap();
            // ENDIF
        }
    }

    /**
     * Perform the restart map operation.
     *
     * <b>preconditions:</b> a map exists and a game is in progress
     * <b>postconditions:</b> the map has been restarted
     */
    public void doRestartMap()
    {
        // CALL class Map method restartLevel
        map.restartLevel();
        // CALL method map.startTimer()
        map.startTimer();
    }

    /**
     * Allows the user to agree or disagree to a restart level request.  If
     * hosting, the restart request is sent to the other users
     *
     * @param nRobot the robot number requesting the change
     *
     * <br><b>precondition:</b> a multiplayer game is in progress
     * <br><b>postcondition:</b> the user has agreed or disagreed to the restart
     * level request
     */
    public void requestRestart(int nRobot)
    {
        // INIT robot = map.getRobot()
        Robot robot = map.getRobot(nRobot);
        // IF hosting THEN
        if (hosting)
        {
            // SET robot's restart flag = true
            robot.setRestartFlag(true);
            // CALL method serv.sendRestartRequest()
            serv.sendRestartRequest(nRobot);
            // INCREMENT restartVotes
            restartVotes++;
            // ENDIF
        }
        // INIT new boolean accepted = showRestartRequestDialog()



        boolean accepted = mui.showRestartRequestDialog(robot.getUserName());
        // IF hosting THEN
        if (hosting)
        {
            // SET robot = loacl robot
            robot = getLocalRobot();
            // CALL method postAcceptedRestart()
            postAcceptedRestart(robot.getRobotNumber(), accepted);
        }
        // ELSE
        else
        {
            // CALL method client.sendRestart accept
            client.sendRestartAccept(accepted);
            // ENDIF
        }
    }

    /**
     * Sets a robot's restart level flag to the accepted boolean passed.
     *
     * @param robotNum the robotnum of the user accepting or declining
     * @param accepted true if the user accepts, false if not
     *
     * <br><b>preconditions:</b> a multiplayer game is in progress
     * <br><b>postconditions:</b> the user with robotNum has selected to accept or
     * decline the request for a level restart
     */
    public void postAcceptedRestart(int robotNum, boolean accepted)
    {
        // IF multiplayer and hosting THEN
        if (multiplayer && hosting)
        {
            // SET robot's restart flag = accepted
            map.getRobot(robotNum).setRestartFlag(accepted);
            // INCREMENT restartVotes
            restartVotes++;
            // CALL method checkRestartVotes()
            checkRestartVotes();
            // ENDIF
        }
    }

    /**
     * Checks to see if the restart level vote has completed.
     *
     * <br><b>preconditions:</b> a restart level vote is in progress
     * <br><b>postconditions:</b> if the vote is complete and it was unanimous
     * in favor then the level is restarted.  otherwise a vote failed message
     * is sent to the chat
     */
    protected void checkRestartVotes()
    {
        // INIT new boolean ready = true
        boolean ready = true;
        // INIT new int index = 0
        int index = 0;
        // INIT robots[] = map.getRobots()
        Robot[] robots = map.getRobots();
        // WHILE ready and index < robots.length DO
        while (ready && index < robots.length)
        {
            // SET ready = robots[index].getRestartFlag()
            ready = robots[index].getRestartFlag();
            // INCREMENT index
            index++;
            // ENDWHILE
        }
        // IF ready THEN



        if (ready)
        {
            // CALL method serv.sendRestartMap()
            serv.sendRestartMap();
            // CALL class Map method restartLevel
            map.restartLevel();
            // CALL method map.startTimer()
            map.startTimer();
            // ENDIF
        }
        // IF restartVotes >= robots.length



        if (restartVotes >= robots.length)
        {
            // FOR bot = 0 to robots length
            for (int bot = 0; bot < robots.length; bot++)
            {
                // SET robots[bot]'s restart flag false
                robots[bot].setRestartFlag(false);
                // ENDFOR
            }
            // SET restartVotes = 0



            restartVotes = 0;
            // ENDIF
        }
    }

    /**
     * Sends instantReplay message to the Map class
     *
     * <BR><B>Precondition:</B> The game is in single player mode.
     * <BR><B>Postcondition:</B> The map returns to the state it was in
     * before the instant replay.
     */
    public void instantReplay()
    {
        // IF not multiplayer THEN
        // CALL class Map method instantReplay
        // ENDIF
        if (!multiplayer)
        {
            map.instantReplay();
        }
    }

    /**
     * Tells the Server to begin listening for connections.
     *
     * <BR><B>Precondition:</B> None.
     * <BR><B>Postcondition:</B> The server is listening for connections.
     */
    public void startServer(int port, MultiplayerDialog multiDiag)
    {
        // IF not multiplayer THEN
        if (!multiplayer)
        {
            // INIT Server serv
            serv = new Server(this, port);
            // CALL serv.start() to start the thread
            serv.start();
            // SET hosting = true
            hosting = true;
            // SET multiplayer = true
            multiplayer = true;
            // SET Client client = null
            client = null;
            // SET mui = md
            mui = multiDiag;
            // SET robotNum = 0
            robotNum = 0;
            // CALL method mui.updateUserList()
            mui.updateUserList(getRobots());
        }
        // ENDIF
    }

    /**
     * Disconnects the client and sends the disconnect message
     * if this game is a client or disconnects all clients connected
     * to the  server if this game is a server.
     *
     * <BR><B>Precondition:</B> Multiplayer is set to true.
     * <BR><B>Postcondition:</B> The game is disconnected from all others.
     */
    public void stopMultiplayer()
    {
        // IF multiplayer THEN
        if (multiplayer)
        {
            // IF hosting THEN
            if (hosting)
            {
                // CALL serv.stopServer()
                serv.stopServer();
            }
            // ELSE
            else
            {
                if (client.isConnected())
                {
                    // CALL client.disconnect()
                    client.disconnect();
                }
                // ENDIF
            }
            // SET hosting = false



            hosting = false;
            // SET multiplayer = false
            multiplayer = false;
            // HIDE mui window
            mui.dispose();
            // ENDIF
        }
    }

    /**
     * Initilizes the client to connect to the server.
     *
     * @param ip The IP address to connect to.
     * @param port The port to use in the connection.
     * <BR><B>Precondition:</B> The game is not running as a host
     * <BR><B>Postcondition:</B> The client is running and the game
     * is set to be a multiplayer game.
     */
    public void startClient(String ip, int port, MultiplayerDialog multiDiag)
    {
        // IF not multiplayer THEN
        if (!multiplayer)
        {
            // REMOVE local robot
            map.removeRobot(robotNum);
            // INIT Client client
            client = new Client(this, ip, port);
            // CALL client.start() to start the thread
            client.start();
            // SET hosting = false
            hosting = false;
            // SET Server serv = null
            serv = null;
            // SET multiplayer = true
            multiplayer = true;
            // SET mui = multiDiag
            mui = multiDiag;
            // ENDIF
        }
    }

    /**
     * Indicates whether or not the game is a multiplayer game.
     */
    public boolean isMultiplayer()
    {
        // RETURN multiplayer
        return multiplayer;
    }

    /**
        * Indicates whether or not the client is connected to a
        * multiplayer game.
        */
    public boolean isHosting()
    {
        // RETURN multiplayer && hosting
        return (multiplayer && hosting);
    }

    /**
        * Indicates whether or not the client is connected to a
        * multiplayer game.
        */
    public boolean isClientConnected()
    {
        if (multiplayer && !hosting)
        {
            // SET multiplayer = multi
            return client.isConnected();
        }
        return false;
    }

    /**
     * Returns the current username
     *
     * @return userName The current name set as the user
     */
    public String getUserName()
    {
        // RETURN userName
        return userName;
    }

    /**
     * Sets the current userName to the passed name
     *
     * @param name the name of a user
     */
    public void setUserName(String name)
    {
        // SET userName = name
        userName = name;
        if (getLocalRobot() != null)
        {
            // CALL method getLocalRobot().setUserName() passing name
            getLocalRobot().setUserName(name);
        }
    }
    /**
     * Returns the current totalScore
     *
     * @return totalScore The current cummulative score for the user
     */
    public int getTotalScore()
    {
        // RETURN totalScore
        return totalScore;
    }

    /**
        * Returns the local Robot
        *
        * @return Returns the Robot associated with the local machine
        */
    public Robot getLocalRobot()
    {
        // IF map != null THEN
        if (map != null)
        {
            // RETURN robotNum
            return map.getRobot(robotNum);
        }
        // ELSE
        else
        {
            // RETURN null
            return null;
        }

    }

    /**
     * Informs the GUI to update the window
     *
     * <br><b>preconditons:</b> none.
     * <br><b>postconditions:</b> the ui display has been repainted
     */
    public void updateDisplay()
    {
        // CALL method gui.validate()
        gui.validate();
        // CALL method gui.repaint()
        gui.repaint();
    }

    /**
     * Informs the multiplayer dialog to display a connection closed
     * dialog message.
     *
     * <br><b>preconditions:</b> multiplayer game is in progress.
     * <br><b>postconditions:</b> the connection closed dialog is displayed
     */
    public void connectionClosed()
    {
        // CALL method mui.connectionClosed()
        mui.connectionClosed();
    }

    /**
     * Returns whether the game is active or not.
     *
     * @return true if the game is active, otherwise false
     */
    public boolean isActive()
    {
        // RETURN active
        return active;
    }

    /**
        * Closes any connection that is present and clears the map.
        *
        * <BR><B>Precondition:</B> The game is running.
        * <BR><B>Postcondition:</B> The game is terminated.
        */
    public void endGame()
    {
        // IF multiplayer THEN
        if (multiplayer)
        {
            // CALL method stopMultiplayer()
            stopMultiplayer();
            // ENDIF
        }
        else
        {
            // ADD score to highscores list
            highScores.addHighScore(new HighScore(userName, totalScore));
        }
        // SET totalScore = 0
        totalScore = 0;
        gui.loadMainMenu();
        updateDisplay();
        active = false;
    }

    /**
        * Returns the cumulative time spent on solving the current map (in seconds).
        *
        * @return Returns the amount of time currently spent solving the active map
        */
    public int getTime()
    {
        // IF debugMode THEN
        if (debugMode)
        {
            // RETURN 1
            return 1;
        }
        // ELSE
        else
        {
            // IF map != null THEN
            if (map != null)
            {
                // RETURN map.getTime()
                return map.getTime();
            }
            // ELSE
            else
            {
                // RETURN 0
                return 0;
            }
        }
    }

    /**
     * Adds an observer to the current map
     *
     * @param obs the Observer to be added to the map
     */
    public void setMapObserver(Observer obs)
    {
        // CALL method Map.deleteObservers()
        map.deleteObservers();
        // CALL method Map.addObserver() passing the obs parameter
        map.addObserver(obs);
    }

    /**
     * Returns a String array containing the save game descriptions.
     *
     * @return String[] continaing the game descriptions.
     */
    public static String[] getSaveDescriptions()
    {
        // INIT new String[] desc
        String[] desc = new String[Constants.maxSaves];
        // CHECK to see if the files exist
        // READ the description line from the save game files
        try
        {
            BufferedReader inFile;
            String filename;
            File saveFile;
            for (int index = 0; index < Constants.maxSaves; index++)
            {
                filename = Constants.savePath + "GAME" + index + Constants.saveExt;
                saveFile = new File(filename);
                if (saveFile.exists())
                {
                    inFile = new BufferedReader(new FileReader(saveFile));
                    desc[index] = inFile.readLine();
                    inFile.close();
                }
                else
                {
                    desc[index] = null;
                }
            }
        }
        catch (IOException exc)
        {
            exc.printStackTrace();
        }
        // RETURN desc
        return desc;
    }

    /**
     * Returns a String array of the names of the levels in the level list.
     *
     * @return a string array with the names of the levels in the level list.
     */
    public String[] getLevelList()
    {
        // RETURN customLevelOrder.getLevelList()
        return customLevelOrder.getLevelList();
    }

    /**
     * Returns the name of the next level.
     *
     * @return the name of the next level or null if the current level is
     * the last level
     */
    public String getNextLevel()
    {
        // RETURN customLevelOrder.getNextLevel()
        return customLevelOrder.getNextLevel();
    }

    /**
     * Returns the name of the previous level.
     *
     * @return the name of the previous level or null if the current level is
     * the first level
     */
    public String getPrevLevel()
    {
        // RETURN customLevelOrder.getPrevLevel()
        return customLevelOrder.getPrevLevel();
    }

    /**
     * Returns the name of the first level in the level list.
     *
     * @return the name of the first level in the level list
     */
    public String getFirstLevel()
    {
        // INIT new String[] levelList
        String[] levelList = customLevelOrder.getLevelList();
        // RETURN the first level name in the levelList
        return levelList[0];
    }

    /**
     * Returns the name of the current level being played
     *
     * @return the name of the current level
     */
    public String getCurrentLevel()
    {
        // RETURN customLevelOrder.getCurrentLevel()
        return customLevelOrder.getCurrentLevel();
    }

    /**
     * Sets whether the getTime() method reports time as always 1 or if
     * it returns the correct value.  debug mode makes time = 1.
     *
     * @param mode turns debug mode on or off
     */
    public void setDebugMode(boolean mode)
    {
        // SET debugMode = mode
        debugMode = mode;
    }
}

