import java.util.*;
import java.io.*;
import javax.swing.*;

/**
 * The HighScoresList class is a list of the top 15 scores achieved during the
 * play of Sokoban.
 *
 * @author Janice Lin
 * @date 11/23/01
 * @version 1.1
 */
//Version History
//  11/23/01 Joshua Thompson added algorithm pseudocode
//	11/28/01 Joshua Thompson modified
//	3/3/02   Antony George implemented the HighScoreList constructor and loadHighScores method
//  3/4/02   Antony George implemented the addHighScore method
//  3/5/02   Antony George implemented the clearHighScores and saveHighScore methods

public class HighScoresList
{
    /**
     * the scores array will hold 15 HighScore objects representing the 15 names and
     * associated scores on the high scores list
     */
    private Vector scores;
    /**
     * The maximum number of scores on the high scores list (15)
     */
    private int maxScores = 15;
    /**
     * Debugging flag for running a JUnit test.
     */
    private boolean junit;
    private Game game;

    /**
     * Constructor - initializes the scores vector.
     */
    public HighScoresList(Game game)
    {
        // INIT scores = new Vector
        scores = new Vector();
        this.game = game;
        this.junit = false;
        loadHighScores();
    }

    /**
     * Constructor - initializes the scores vector.
     */
    public HighScoresList(boolean junit)
    {
        // INIT scores = new Vector
        scores = new Vector();
        this.game = null;
        this.junit = junit;
        loadHighScores();
    }

    /**
     * The loadHighScores method allows the user to load the scores from a file.
     * <p>
     * <BR><B>Precondition:</B> high scores file exists
     * <BR><B>Postcondition:</B>The list of high scores is loaded to be viewed
     * by the user
     */
    public void loadHighScores()
    {
        try
        {   
            // INIT File scores.dat
            File scoreData = new File("scores.dat"); 
            // IF the file doesn't exist
            if (!scoreData.exists())
            {
                // CREATE the file
                try
                {
                    BufferedWriter outFile = new BufferedWriter(new FileWriter(scoreData));
                    outFile.close();
                }
                catch (IOException err2)
                {
                }
            }
            // OPEN file to be read

            BufferedReader in = new BufferedReader(new FileReader(scoreData));
            // CLEAR the current high scores
            scores.removeAllElements();
            String input;
            // WHILE the input stream is ready
            while ((input = in.readLine()) != null)
            {
                // READ line in file
                StringTokenizer tokens = new StringTokenizer(input, " ");
                String name = tokens.nextToken();
                int score = Integer.parseInt(tokens.nextToken());
                // CREATE new HighScore with name and score on line
                HighScore newscore = new HighScore(name, score);
                // ADD new HighScore to corresponding location in vector
                scores.add(newscore);
                // ENDWHILE
            }
            // CLOSE file


            in.close();
        }
        catch(IOException err)
        {}
    }

    /**
     * The addScores method allows the user to add scores to the high scores list.
     * It finds where the score belongs and then inserts the score at the
     * appropriate place.  If there are more than 15 users on the high scores
     * list then the method removes the last score from the list.
     *
     * @param score a HighScore class with name and score to be added to the high 
     * scores list
     * <p>
     * <BR><B>Precondition:</B>Checks to see if the file have space for name and score entry
     * <BR><B>Postcondition:</B>The added name and score will be saved in the high
     * scores list.
     *
     */
    public void addHighScore(HighScore score)
    {
        loadHighScores();
        // SET location to 1
        int location = scores.size();
        // SET validScore = false
        boolean validScore = false;
        // CREATE new HighScore with name and score passed
        // FOR index = 0 to end of scores list
        int index = 0;
        while (index < scores.size() && !validScore)
        {
            // IF score is greater than scores[index] THEN
            if (score.getScore() > ((HighScore)scores.get(index)).getScore())
            {
                // SET location to index
                location = index;
                // SET validScore to true
                validScore = true;
                // ENDIF
            }
            index++;
            // ENDWHILE
        }
        if (!junit && (validScore || (scores.size() == 0 && score.getScore() > 0)))
        {
            String scoreTell = new String("You have a high score! Please enter your name:");
            String title = "New High Score";
            String newname = JOptionPane.showInputDialog(game.gui, 
                                                  scoreTell,
                                                  title,
                                                  JOptionPane.PLAIN_MESSAGE);
            if (newname != null && !newname.equals(""))
            {
                score.setName(newname);
            }
        }
        // remove all spaces from name
        StringTokenizer tkns = new StringTokenizer(score.getName(), " ");
        String fix = "";
        while (tkns.hasMoreTokens())
        {
            fix = fix + tkns.nextToken();
        }
        // set score name to name without spaces
        score.setName(fix);
        // IF validScore is true THEN
        if (validScore || scores.size() < maxScores)
        {
            if (score.getScore() > 0)
            {                // ADD score to Vector at location
                scores.add(location, score);
            }
            // IF number of scores is greater than maxScores THEN
            if (scores.size() > maxScores)
            {
                // DELETE last object on scores
                scores.removeElementAt(scores.size()-1);
                // ENDIF
            }
            // ENDIF
        }
        // CALL method saveHighScores()

        this.saveHighScores();
    }

    /**
     * The clearHighScores method allows the user to clear the current list of high
     * scores.
     *
     * <BR><B>Precondition:</B>None
     * <BR><B>Postcondition:</B>high scores list is empty
     */
    public void clearHighScores()
    {
        // SET scores = new empty Vector
        scores = new Vector();
        // CALL method saveHighScores()
        this.saveHighScores();
    }


    /**
     * The saveScores method allows the user to save the current list of high
     * scores to a file.
     *
     * <BR><B>Precondition:</B>User has completed a level
     * <BR><B>Postcondition:</B>The score is saved
     */
    private void saveHighScores()
    {
        try
        {
            // OPEN file to be written
            BufferedWriter outFile = new BufferedWriter(new FileWriter("scores.dat"));
            // FOR index = 0 to end of scores list
            for (int index = 0; index < scores.size(); index++)
            {
                // WRITE name and score to a line in file
                HighScore output = (HighScore)scores.get(index);
                outFile.write(output.getName() + " " + output.getScore());
                outFile.newLine();
                // ENDFOR
            }
            // CLOSE file


            outFile.close();
        }
        catch(IOException err)
        {
            System.out.println("IO Exception");
        }
    }

    /**
     * Returns the High Scores List as an array of HighScores.
     */
    protected HighScore[] toArray()
    {
        Object[] objarray = scores.toArray();
        HighScore[] retarray = new HighScore[objarray.length];
        for (int index = 0; index < objarray.length; index++)
        {
            retarray[index] = (HighScore)objarray[index];
        }
        return retarray;
    }
}


