import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
  * The MapRenderer class is a graphical representation of the current state of the map.
  *
  * @author  Joshua Thompson
  * @date    12/01/01
  * @version 1.1
  *
  */
//Version History
//  12/01/01 Joshua Thompson added java code
//  01/28/02 Joshua Thompson added java code for implementing MapRenderer as an
//			 Observer of the Map class

public class MapRenderer extends JPanel implements Observer
{
    /**
     * A pointer to the main Game object
     */
    private Game owner;
    /**
     * Contains the GIF file that represents a robot facing up.
     */
    private Image robotUp = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"robotup.gif");
    /**
     * Contains the GIF file that represents a robot facing down.
     */
    private Image robotDown = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"robotdown.gif");
    /**
     * Contains the GIF file that represents a robot facing left.
     */
    private Image robotLeft = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"robotleft.gif");
    /**
     * Contains the GIF file that represents a robot facing right.
     */
    private Image robotRight = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"robotright.gif");
    /**
     * Contains the GIF file that represents a crate.
     */
    private Image crate = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"crate.gif");
    /**
     * Contains the GIF file that represents an empty square.
     */
    private Image empty = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"empty.gif");
    /**
     * Contains the GIF file that represents an incinerator.
     */
    private Image incinerator = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"incinerator.gif");
    /**
     * Contains the GIF file that represents a wall.
     */
    private Image wall = Toolkit.getDefaultToolkit().getImage(Constants.graphicsPath+"wall.gif");
    /**
     * Whether or not all images have been loaded (used to prevent initial flashing robot).
     */
    private boolean loaded = false;

    /**
     * Constructor of MapRenderer - inits the point to the Game class and sets the size
     * and the background color of the JPanel.
     */
    public MapRenderer ()
    {
        // CALL method MapRenderer.setSize() passing 500 and 480
        this.setSize(500, 480);
        // CALL method MapRenderer.setBackground() passing Color.black
        this.setBackground(Color.black);
    }

    /**
     * Set the link back to the Game class so that the maprenderer can
     * get the map to know what to update.
     *
     * @param owner the instance of Game where MapRenderer will get the map
     */
    public void setOwner (Game owner)
    {
        this.owner = owner;
    }

    /**
     * Paints the 9 squares surrounding the point passed to the method as the arg
     * parameter.  If arg is null then the method paintComponent is called to repaint
     * the entire map instead of one section.
     *
     * @param o the observable map class
     * @param arg the argument passed from the notifyObservers method
     * <p><b>Pre-conditions:</b>the maprenderer is currently observing a map.</p>
     * <p><b>Post-conditions:</b>the display of the map has been updated to reflect
     * the change.</p>
     */
    public void update(Observable o, Object arg)
    {
        // IF owner != null THEN
        if (owner != null)
        {
            // GET the graphics context
            Graphics gc = this.getGraphics();
            // IF arg = null THEN
            if (arg == null)
            {
                // REDRAW the whole map
                this.paintComponent(gc);
            }
            // ELSE
            else
            {
                // SET Point pt = arg
                Point pt = (Point)arg;
                // INIT new Square[][] = returned value from owner.getMap()
                Square[][] map = owner.getMap();
                // SET Robot robot = returned value from owner.getLocalRobot()
                Robot robot = owner.getLocalRobot();
                // INIT new int row = the y value of the Point pt
                int row = pt.y - 1;
                // INIT new int col = the x value of the Point pt
                int col = pt.x - 1;
                // DECLARE x and y
                int x;
                int y;
                // INIT new Image img = an empty image
                Image img = empty;
                // DRAW the 9 squares surrounding the robot
                for (int mrow = row; mrow <= (pt.y + 1); mrow++)
                {
                    col = pt.x - 1;
                    for (int mcol = col; mcol <= (pt.x + 1); mcol++)
                    {
                        row = mrow;
                        col = mcol;
                        if (map[row][col] == Square.EMPTY)
                        {
                            img = empty;
                        }
                        else if (map[row][col] == Square.CRATE)
                        {
                            img = crate;
                        }
                        else if (map[row][col] == Square.WALL)
                        {
                            img = wall;
                        }
                        else if (map[row][col] == Square.INCINERATOR)
                        {
                            img = incinerator;
                        }
                        else if (map[row][col] == Square.ROBOT)
                        {
                            if (robot.getDirectionFaced() == Direction.UP)
                            {
                                img = robotUp;
                            }
                            else if (robot.getDirectionFaced() == Direction.DOWN)
                            {
                                img = robotDown;
                            }
                            else if (robot.getDirectionFaced() == Direction.LEFT)
                            {
                                img = robotLeft;
                            }
                            else if (robot.getDirectionFaced() == Direction.RIGHT)
                            {
                                img = robotRight;
                            }
                        }
                        y = row*Constants.gSizeY;
                        x = col*Constants.gSizeX;
                        gc.drawImage(img, x, y, this);
                        // ENDFOR
                    }
                    // ENDFOR
                }
                // ENDIF
            }
            // DISPOSE of the graphics context

            gc.dispose();
            // ENDIF
        }
    }

    /**
     * Paints the JPanel, drawing the necessary images to present an accurate graphical
     * representation of the current map.
     *
     * @param g the graphics object used for drawing images.
     * <p><b>Pre-conditions:</b>none.</p>
     * <p><b>Post-conditions:</b>the current map has been displayed.</p>
     */
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);
        // IF the robot images haven't been loaded THEN
        if (!loaded)
        {
            // DRAW robot images up, down, left and right to prevent flashing robot
            g.drawImage(robotUp, 0, 0, this);
            g.drawImage(robotDown, 0, 0, this);
            g.drawImage(robotLeft, 0, 0, this);
            g.drawImage(robotRight, 0, 0, this);
            g.drawImage(empty, 0, 0, this);
            // SET loaded = true
            loaded = true;
            // ENDIF
        }
        // GET the map from the game class calling owner.getMap()

        Square[][] map = owner.getMap();
        // IF map = null THEN
        if (map != null)
        {
            // INIT Image img as temporary storage
            Image img = empty;
            // INIT x and y variables to 0
            int x;
            int y;
            // DRAW the images for the board
            for (int row = 0; row < Constants.maxRow; row++)
            {
                for (int col = 0; col < Constants.maxCol; col++)
                {
                    if (map[row][col] == Square.EMPTY)
                    {
                        img = empty;
                    }
                    else if (map[row][col] == Square.CRATE)
                    {
                        img = crate;
                    }
                    else if (map[row][col] == Square.WALL)
                    {
                        img = wall;
                    }
                    else if (map[row][col] == Square.INCINERATOR)
                    {
                        img = incinerator;
                    }
                    x = col*Constants.gSizeX;
                    y = row*Constants.gSizeY;
                    g.drawImage(img, x, y, this);
                }
            }

            // CALL method owner.getRobots()
            // INIT new Robot[] robots to returned array from owner.getRobots()
            Robot[] robots = owner.getRobots();
            // Draw the images for the robots
            for (int index = 0; index < robots.length; index++)
            {
                if (robots[index].getDirectionFaced() == Direction.UP)
                {
                    img = robotUp;
                }
                else if (robots[index].getDirectionFaced() == Direction.DOWN)
                {
                    img = robotDown;
                }
                else if (robots[index].getDirectionFaced() == Direction.LEFT)
                {
                    img = robotLeft;
                }
                else if (robots[index].getDirectionFaced() == Direction.RIGHT)
                {
                    img = robotRight;
                }
                x = robots[index].getCol()*Constants.gSizeX;
                y = robots[index].getRow()*Constants.gSizeY;
                g.drawImage(img, x, y, this);
            }
        }
    }
}
