import java.util.*;

/**
* The robot class represents an individual robot that is active on
* a level at any given time. A different robot should be created for
* each player in the game.
*
* @author 	Janice Lin
* @date	11/23/01
* @version 1.3
*
*/
//Version History
//  11/23/01 Janice Lin added algorithm pseudocode
//  11/28/01 Janice Lin modified javadoc headers and added variable
//			 descriptions
//  01/11/02 Joshua Thompson seperated moveTracker into new class
//	01/12/02 Joshua Thompson modified java code to use new MoveTracker
//			 class

public class Robot
{
    /**
     * The color of the robot
     */
    private RobotColor color;
    /**
     * The robot identification number
     */
    private int robotNumber;
    /**
     * The userName of the robot
     */
    private String userName;
    /**
     * The row position of the robot on the map
     */
    private int row;
    /**
     * The column position of the robot on the map
     */
    private int col;
    /**
     * The total score for the current level
     */
    private int score;
    /**
     * The number of times a robot move has resulted in pushing a crate
     */
    private int pushes;
    /**
     * The number of times a robot has moved
     */
    private int moves;
    /**
     * The number of times a push has resulted in a crate being incinerated
     */
    private int cratesIncinerated;
    /**
     * The list of all of the moves a robot has made
     */
    private MoveTracker moveTracker;

    /**
     * The direction the robot is currently facing
     */
    private Direction directionFaced;

    /**
     * The flag signifying if the robot wishes to restart the level.
     */
    private boolean restart;

    /**
     * The flag signifying if the robot wishes to change the level.
     */
    private boolean changeLevel;

    /**
     * Creates a new Robot object of the given color at the given
     * position and sets all other variables to their initial conditions.
     *
     * @param userName the name of the robot
     * @param robotNum	The number of the robot
     */
    public Robot(String userName, int robotNum)
    {
        // SET robotNumber = robotNum
        // SET robotColor = color
        // SET moves = 0
        // SET pushes = 0
        // SET cratesIncinerated = 0
        // INIT moveTracker Vector
        this.userName = userName;
        this.robotNumber = robotNum;
        this.color = RobotColor.BLUE;
        this.directionFaced = Direction.RIGHT;
        this.restart = false;
        this.changeLevel = false;
        moves = 0;
        pushes = 0;
        cratesIncinerated = 0;
        moveTracker = new MoveTracker();
    }

    /**
     * Returns the current color of the robot
     *
     * @return The enumerated RobotColor associated with the color of the robot.
     */
    public RobotColor getColor()
    {
        //RETURN robotColor
        return color;
    }

    /**
     * Sets the current color of the robot.
     *
     * @param color	The new int value for color.
     */
    public void setColor(RobotColor color)
    {
        //SET robotColor = color;
        this.color = color;
    }

    /**
     * Returns which robot this is.
     *
     * @return The int associated with the color
     */
    public int getRobotNumber()
    {
        //RETURN robotNumber
        return robotNumber;
    }

    /**
     * Returns the current row location of the robot.
     *
     * @return The int value representing the row location
     */
    public int getRow()
    {
        //RETURN row
        return row;
    }

    /**
     * Sets the current row location of the robot.
     *
     * @param row	The new int value representing the row location
     */
    public void setRow(int row)
    {
        //SET row = passed row;
        this.row = row;
    }

    /**
     * Returns the current column location of the robot.
     *
     * @return The int value representing the column location
     */
    public int getCol()
    {
        //RETURN col
        return col;
    }

    /**
     * Sets the current column location of the robot.
     *
     * @param column	The int value representing the column location
     */
    public void setCol(int column)
    {
        //SET col = column;
        col = column;
    }

    /**
      * Adds the inputed move to the Robot's moveTracker and adjusts the
      * robots location.
      *
      * @param mv	The move to be added to the vector
      * @param blocked true if the robot is blocked, otherwise false
      * <BR><B>Precondition:</b> None
      * <BR><B>Postcondition:</B> The move was added to the moveTracer Vector
      */
    public void addMove(Move mv)
    {
        // CALL method moveTracker.add() passing mv
        moveTracker.add(mv);
        // INIT new Direction dir to value returned by mv.getDirection()
        Direction dir = mv.getDirection();
        // IF the robot has moved THEN
        if (mv.hasRobotMoved())
        {
            // INCREMENT moves
            moves++;
            // CASE dir OF
            // UP: INCREMENT row
            // DOWN: DECREMENT row
            // LEFT: DECREMENT col
            // RIGHT: INCREMENT col
            // ENDCASE
            // ADJUST the row and col of the robot
            if(dir == Direction.UP)
            {
                row--;
            }
            else if(dir == Direction.DOWN)
            {
                row++;
            }
            else if(dir == Direction.LEFT)
            {
                col--;
            }
            else if(dir == Direction.RIGHT)
            {
                col++;
            }
            // IF mv.hasCrateMoved() is true THEN
            // INCREMENT pushes
            // ENDIF
            if(mv.hasCrateMoved())
            {
                pushes++;
            }
            // IF mv.wasIncinerated() is true THEN
            // INCREMENT cratesIncinerated
            // ENDIF
            if(mv.wasIncinerated())
            {
                cratesIncinerated++;
            }
            // ENDIF
        }
        // SET directionFaced = dir;

        directionFaced = dir;
    }

    /**
     * Removes a move from the moveTracker Vector.
     *
     * @return	Returns the move that was removed or null if the moveTracker is
     * empty
     *<BR><B>Precondition:</B> There needs to be a move on the moveTracker vector
     *<BR><B>Postcondition:</B> The move was removed from the vector
     */
    public Move removeMove()
    {
        // IF moveTracker.isEmpty() returns false THEN
        if(!moveTracker.isEmpty())
        {
            // CALL moveTracker.removeLast()
            // INIT new Move to returned Move from moveTracker.removeLast()
            Move removedmove = moveTracker.removeLast();
            // IF the robot has moved THEN
            if (removedmove.hasRobotMoved())
            {
                // DECREMENT moves
                moves--;
                // ADJUST row and column to reflect new robot location
                Direction direction = removedmove.getDirection();
                if (direction == Direction.UP)
                {
                    row++;
                }
                else if (direction == Direction.DOWN)
                {
                    row--;
                }
                else if (direction == Direction.LEFT)
                {
                    col++;
                }
                else if (direction == Direction.RIGHT)
                {
                    col--;
                }
                // IF the returned Move.hasCrateMoved() is true THEN
                // DECREMENT pushes
                // ENDIF
                if(removedmove.hasCrateMoved())
                {
                    pushes--;
                }
                // IF the returned Move.wasIncinerated() is true THEN
                // DECREMENT cratesIncinerated
                // ENDIF
                if(removedmove.wasIncinerated())
                {
                    cratesIncinerated--;
                }
                // ENDIF
            }
            // SET directionFaced to the new last move

            Move lastmove = moveTracker.lastElement();
            // IF lastmove != null THEN
            if (lastmove != null)
            {
                // SET directionFaced = direction of last move
                directionFaced = lastmove.getDirection();
            }
            else
            {
                // SET directionFaced = Direction.RIGHT
                directionFaced = Direction.RIGHT;
            }
            // RETURN Move
            return removedmove;
        }
        // ELSE
        else
        {
            // RETURN null
            return null;
            // ENDIF
        }
    }

    /**
      * Resets the moveTracker to an empty value.
      *<BR><B>Precondition:</B> None
      *<BR><B>Postcondition:</B> The moveTracker vector is set to empty
      */
    public void resetMoveTracker()
    {
        // CALL moveTracker.clear()
        moveTracker.clear();
        // SET moves = 0
        moves = 0;
        // SET pushes = 0
        pushes = 0;
        // SET cratesIncinerated = 0
        cratesIncinerated = 0;
        // SET directionFaced = Direction.RIGHT
        directionFaced = Direction.RIGHT;
    }

    /**
     * Returns an array of the moves within the MoveTracker
     *
     * @return an array of Move
     * <BR><B>Precondition:</B> moveTracker is not empty
     * <BR><B>Postcondition:</B> an array of Move is returned
     */
    public Move[] getMoveTracker()
    {
        // RETURN moveTracker.getMoveTracker()
        return moveTracker.getMoveTracker();
    }

    /**
      * Returns the number of moves that have been taken by this robot in the
      * current game.
      *
      * @return 	The number of moves that have been taken
      */
    public int getMoves()
    {
        //RETURN moves
        return moves;
    }

    /**
     * Returns the current score associated with this robot.
     *
     * @return 	The current score
     */
    public int getScore()
    {
        //RETURN score
        return score;
    }

    /**
     * Calculates and returns the score.
     *
     * @param time		Total time has elapsed
      * <BR><B>Precondition:</B> none
     * <BR><B>Postcondition:</B> score is set to the calculated value and
     * returned
     * @return score - the new score
     */
    public int calcScore(int time)
    {
        // score = (cratesIncinerated * 100 - moves )/ time
        score = (cratesIncinerated * 100 - moves)/time;
        // RETURN score
        return score;
    }

    /**
     * Returns the number of pushes that this robot has done in the
     * current game.
     *
     * @return 	The number of pushes that this robot has done
     */
    public int getPushes()
    {
        //RETURN pushes
        return pushes;
    }

    /**
     * Sets the number of moves that the robot has taken.
     *
     * @param pushes 	The new number of pushes done
     */
    public void resetPushes()
    {
        //SET pushes = 0
        pushes = 0;
    }

    /**
     * Returns the number of crates that have been incinerated by this robot in the
     * current game.
     *
     * @return 	The number of crates that have been incinerated
     */
    public int getCratesIncinerated()
    {
        //RETURN cratesIncinerated
        return cratesIncinerated;
    }

    /**
     * Sets the number of crates that the robot has incinerated.
     *
     * @param crates 	The new number of crates that the robot has incinerated
     */
    public void resetCratesIncinerated()
    {
        //SET cratesIncinerated = 0
        cratesIncinerated = 0;
    }

    /**
     * Returns the next color in the color order: blue, red, green and yellow.
     * The order wraps around so that the next color after yellow is blue.
     * <BR><B>Precondition:</B> none
     * <BR><B>Postcondition:</B> color is set to the next color in the order
     * of colors
     */
    public void setNextColor()
    {
        // CASE color OF
        // BLUE: color = RobotColor.RED
        // RED: color = RobotColor.GREEN
        // GREEN: color = RobotColor.YELLOW
        // YELLOW: color = RobotColor.BLUE
        // ENDCASE
        if(color == RobotColor.BLUE)
            color = RobotColor.RED;
        else if(color == RobotColor.RED)
            color = RobotColor.GREEN;
        else if(color == RobotColor.GREEN)
            color = RobotColor.YELLOW;
        else if(color == RobotColor.YELLOW)
            color = RobotColor.BLUE;
    }

    /**
     * Returns the user name of the robot.
     *
     * @return Returns the user name for this robot
     */
    public String getUserName()
    {
        return userName;
    }

    /**
     * Sets the user name of the robot.
     *
     * @param name the new user name for this robot
     */
    public void setUserName(String name)
    {
        // SET userName = name;
        this.userName = name;
    }

    /**
      * Returns the direction that the robot is facing in the
      * current game.
      *
      * @return 	The direction the robot is facing
      */
    public Direction getDirectionFaced()
    {
        //RETURN pushes
        return directionFaced;
    }

    /**
     * Returns the restart flag for the robot signifying if the user
     * wants to restart the level or not.
     *
     * @return Returns the restart flag for the robot
     */
    public boolean getRestartFlag()
    {
        return restart;
    }

    /**
     * Sets the restart flag for the robot.
     *
     * @param flag the flag signifying if the robot wants to restart the
     * level or not
     */
    public void setRestartFlag(boolean flag)
    {
        // SET restart = flag
        restart = flag;
    }

    /**
     * Returns the changeLevel flag for the robot signifying if the user
     * wants to change the level or not.
     *
     * @return Returns the changeLevel flag for the robot
     */
    public boolean getChangeLevelFlag()
    {
        return changeLevel;
    }

    /**
     * Sets the changeLevel flag for the robot.
     *
     * @param flag the flag signifying if the robot wants to change the
     * level or not
     */
    public void setChangeLevelFlag(boolean flag)
    {
        // SET changeLevel = flag
        changeLevel = flag;
    }
}
