import java.io.*;
import java.net.*;
import java.util.*;

/**
 * <p>The ServerThread class is a communications interface between the server
 * and the client.</p>
 *
 * @author Joshua Thompson
 * @date 11/23/2001
 * @version 1.2
 *
 */
//Version History
//  11/23/01 Joshua Thompson added algorithm pseudocode
//  11/28/01 Joshua Thompson modified javadoc headers and added variable
//			 descriptions

public class ServerThread extends Thread
{
    /**
     * a pointer to the main Game object
     */
    private Game owner;
    /**
     * the socket connection to the client
     */
    private Socket socket;
    /**
     * the output stream to the client
     */
    private PrintWriter out;
    /**
     * the input stream from the client
     */
    private BufferedReader in;
    /**
     * the thread number used for identification purposes
     */
    private int threadNum;
    /**
     * the user name of the client connected
     */
    private String userName;
    /**
     * reading is true when accepting commands from the client and false when not
     */
    private boolean reading;

    /**
     * <p>Initializes ServerThread, receives input from clients and closes
     * the connection when finished.</p>
     *
     * <p><b>Pre-condition:</b> existing socket connection to a client </p>
     * <p><b>Post-condition:</b> data has been received and socket is closed</p>
     */
    public void run()
    {
        try
        {
            // INIT PrintWriter out
            this.out = new PrintWriter(socket.getOutputStream(), true);
            // INIT BufferedReader in
            this.in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
            // WRITE welcome message to PrintWriter out
            out.println("Welcome: Sokoban");
            // SET userName to name read in from BufferedReader in

            userName = in.readLine();
            // WRITE threadNum to PrintWriter out
            out.println(threadNum);
            // SEND current userList to client
            Robot[] robots = owner.getRobots();
            for (int index = 0; index < robots.length; index++)
            {
                out.println("adduser "+robots[index].getUserName()+" "+robots[index].getRobotNumber());
            }
            // CALL method owner.postNewUser() passing userName and threadNum
            owner.postNewUser(userName, threadNum);
            // SET reading true
            reading = true;
            String command;
            // WHILE reading LOOP
            while (reading)
            {
                // READ command from BufferedReader in
                command = in.readLine();
                // CALL method processCommand() passing the command read in
                this.processCommand(command);
                // ENDWHILE
            }
            // CLOSE PrintWriter out

            out.close();
            // CLOSE BufferedReader in
            in.close();
            // CLOSE socket
            socket.close();
        }
        catch (IOException exc)
        {
            //exc.printStackTrace();
            owner.postDisconnect(threadNum);
        }
    }

    /**
     * <p>Sends the command parameter to the client.</p>
     *
     * <p><b>Pre-condition:</b> existing socket connection to a client</p>
     * <p><b>Post-condition:</b> command parameter is sent to the client</p>
     *
     * @param command information to be sent to the client
     */
    public void sendCommand(String command)
    {
        // WRITE command to socket using PrintWriter out
        out.println(command);
    }

    /**
     * <p>Checks the command received from the client and determines what to
     * do with the information</p>
     *
     * <p><b>Pre-condition:</b> command contains the input sent from the client</p>
     * <p><b>Post-condition:</b> the input has been processed and the Game is
     * updated</p>
     *
     * @param command the String recieved from the client to be processed
     */
    public void processCommand(String command)
    {
        StringTokenizer commands = new StringTokenizer(command, " ");
        String commanded = commands.nextToken();
        // CASE command OF
        // move:
        if (commanded.equals("move"))
        {
            doMove(commands);
        }
        // undo:
        else if (commanded.equals("undo"))
        {
            doUndo(commands);
        }
        // chat:
        else if (commanded.equals("chat"))
        {
            doChat(commands);
        }
        // restart request:
        else if (commanded.equals("resreq"))
        {
            owner.requestRestart(threadNum);
        }
        // restart accepted:
        else if (commanded.equals("restart"))
        {
            doRestart(commands);
        }
        // change level request:
        else if (commanded.equals("chgreq"))
        {
            doChangeRequest(commands);
        }
        // change level accept:
        else if (commanded.equals("change"))
        {
            doAcceptChange(commands);
        }
        // disconnect:
        else if (commanded.equals("disconnect"))
        {
            doDisconnect(commands);
            // ENDCASE
        }
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doMove(StringTokenizer commands)
    {
        // READ robotNumber
        int robotNumber = Integer.parseInt(commands.nextToken());
        // READ direction
        if (robotNumber == threadNum)
        {
            String dirStr = commands.nextToken();
            Direction dir = Direction.RIGHT;
            if (dirStr.equals("UP"))
            {
                dir = Direction.UP;
            }
            else if (dirStr.equals("DOWN"))
            {
                dir = Direction.DOWN;
            }
            else if (dirStr.equals("LEFT"))
            {
                dir = Direction.LEFT;
            }
            // CALL method postMove() from owner
            owner.postMove(robotNumber, dir);
        }
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doUndo(StringTokenizer commands)
    {
        // READ robotNumber
        int robotNumber = Integer.parseInt(commands.nextToken());
        if (robotNumber == threadNum)
        {
            // CALL method undoMove() from owner
            owner.undoMove(robotNumber);
        }
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doChat(StringTokenizer commands)
    {
        // READ user name
        String name = commands.nextToken();
        String message = new String("");
        if (name.equals(userName))
        {
            // READ message
            while(commands.hasMoreTokens())
            {
                message = message + commands.nextToken() + " ";
            }
            // CALL method postChat() from owner passing user name and message
            owner.postChat(name, message);
        }
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doRestart(StringTokenizer commands)
    {
        boolean accepted = commands.nextToken().equals("true");
        owner.postAcceptedRestart(threadNum, accepted);
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doChangeRequest(StringTokenizer commands)
    {
        String level = commands.nextToken("");
        level = level.substring(1, level.length());
        owner.requestChangeLevel(level, threadNum);
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doAcceptChange(StringTokenizer commands)
    {
        boolean accepted = commands.nextToken().equals("true");
        String level = commands.nextToken("");
        level = level.substring(1, level.length());
        owner.postAcceptedChangeLevel(level, threadNum, accepted);
    }

    /**
     * <p>Used by processCommand()</p>
     *
     * @param command the String recieved from process to be processed
     */
    private void doDisconnect(StringTokenizer commands)
    {
        // SET reading false
        reading = false;
        // CALL method postDisconnect() from owner passing robot number
        owner.postDisconnect(threadNum);
    }

    /**
     * <p>ServerThread constructor</p>
     *
     * <p><b>Pre-condition:</b> none</p>
     * <p><b>Post-condition:</b> ServerThread class created; owner, socket and
     * threadNum are set to values passed</p>
     *
     * @param owner provides a link back to the game
     * @param socket the Socket connection to the client
     * @param threadNum an identification number signifying which ServerThread it is
     */
    public ServerThread(Game owner, Socket socket, int threadNum)
    {
        // SET owner = passed owner
        this.owner = owner;
        // SET socket = passed socket
        this.socket = socket;
        // SET threadNum = passed threadNum
        this.threadNum = threadNum;
        // SET userName = passed userName
        this.userName = userName;
    }

    /**
     * <p>Returns the thread number of this connection.</p>
     *
     * @return threadNum - the thread number of this connection
     */
    public int getThreadNum()
    {
        // RETURN threadNum
        return threadNum;
    }

    /**
     * <p>Returns the user name of the client connected.</p>
     *
     * @return userName - the user name of the client connected
     */
    public String getUserName()
    {
        // RETURN userName
        return userName;
    }
}
